# -*- Mode: python; coding: utf-8; tab-width: 8; indent-tabs-mode: t; -*-
#
# Copyright (C) 2006 Adam Zimmerman  <adam_zimmerman@sfu.ca>
# Copyright (C) 2006 James Livingston  <doclivingston@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The Rhythmbox authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and Rhythmbox. This permission is above and beyond the permissions granted
# by the GPL license by which Rhythmbox is covered. If you modify this code
# you may extend this exception to your version of the code, but you are not
# obligated to do so. If you do not wish to do so, delete this exception
# statement from your version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.

import zipfile
import sys, os.path
import xml
import datetime
import string
import re

import rb
from gi.repository import RB
from gi.repository import GObject, Gtk, Gio, Peas, PeasGtk

from MagnatuneSource import MagnatuneSource
import MagnatuneAccount

import gettext
gettext.install('rhythmbox', RB.locale_dir())

class MagnatuneEntryType(RB.RhythmDBEntryType):
	def __init__(self):
		RB.RhythmDBEntryType.__init__(self, name='magnatune')
		self.URIre = re.compile(r'^http://[^.]+\.magnatune\.com/')
		self.nsre = re.compile(r'\.(mp3|ogg)$')
		self.account = MagnatuneAccount.instance()

	def fix_trackurl(self, trackurl, account_type, username, password):
		return trackurl

	def do_get_playback_uri(self, entry):
		(account_type, username, password) = self.account.get()
		uri = entry.get_string(RB.RhythmDBPropType.LOCATION)
		if account_type != "none":
			uri = self.URIre.sub("http://%s:%s@%s.magnatune.com/" % (username, password, account_type), uri)
			uri = self.nsre.sub(r"_nospeech.\1", uri)
			print("converted track uri: %s" % uri)

		return uri

	def do_can_sync_metadata(self, entry):
		return True

	def do_sync_metadata(self, entry, changes):
		return



class Magnatune(GObject.GObject, Peas.Activatable):
	__gtype_name__ = 'Magnatune'
	object = GObject.property(type=GObject.GObject)

	def __init__(self):
		GObject.GObject.__init__(self)

	def download_album_action_cb(self, action, parameter):
		shell = self.object
		shell.props.selected_page.download_album()

	def artist_info_action_cb(self, action, parameter):
		shell = self.object
		shell.props.selected_page.display_artist_info()
	
	def do_activate(self):
		shell = self.object
		self.db = shell.props.db

		rb.append_plugin_source_path(self, "icons")

		self.entry_type = MagnatuneEntryType()
		self.db.register_entry_type(self.entry_type)

		self.settings = Gio.Settings.new("org.gnome.rhythmbox.plugins.magnatune")

		app = Gio.Application.get_default()
		action = Gio.SimpleAction(name="magnatune-album-download")
		action.connect("activate", self.download_album_action_cb)
		app.add_action(action)

		action = Gio.SimpleAction(name="magnatune-artist-info")
		action.connect("activate", self.artist_info_action_cb)
		app.add_action(action)

		builder = Gtk.Builder()
		builder.add_from_file(rb.find_plugin_file(self, "magnatune-toolbar.ui"))
		toolbar = builder.get_object("magnatune-toolbar")
		app.link_shared_menus(toolbar)

		group = RB.DisplayPageGroup.get_by_id ("stores")
		settings = Gio.Settings.new("org.gnome.rhythmbox.plugins.magnatune")
		self.source = GObject.new(MagnatuneSource,
					  shell=shell,
					  entry_type=self.entry_type,
					  icon=Gio.ThemedIcon.new("magnatune-symbolic"),
					  plugin=self,
					  settings=settings.get_child("source"),
					  name=_("Magnatune"),
					  toolbar_menu=toolbar)

		shell.register_entry_type_for_source(self.source, self.entry_type)
		shell.append_display_page(self.source, group)

		self.pec_id = shell.props.shell_player.connect('playing-song-changed', self.playing_entry_changed)

	def do_deactivate(self):
		shell = self.object

		shell.props.shell_player.disconnect(self.pec_id)

		self.db.entry_delete_by_type(self.entry_type)
		self.db.commit()
		self.db = None
		self.entry_type = None
		self.source.delete_thyself()
		self.source = None

	def playing_entry_changed (self, sp, entry):
		self.source.playing_entry_changed(entry)




class MagnatuneConfig(GObject.GObject, PeasGtk.Configurable):
	__gtype_name__ = 'MagnatuneConfig'
	object = GObject.property(type=GObject.GObject)

	format_list = ['ogg', 'flac', 'wav', 'mp3-vbr', 'mp3-cbr']

	def __init__(self):
		GObject.GObject.__init__(self)
		self.settings = Gio.Settings.new("org.gnome.rhythmbox.plugins.magnatune")
		self.account = MagnatuneAccount.instance()

	def do_create_configure_widget(self):
		# We use a dictionary so we can modify these values from within inner functions
		keyring_data = {
			'id': 0,
			'item': None
		}

		def update_sensitivity(account_type):
			has_account = account_type != "none"
			builder.get_object("username_entry").set_sensitive(has_account)
			builder.get_object("password_entry").set_sensitive(has_account)
			builder.get_object("username_label").set_sensitive(has_account)
			builder.get_object("password_label").set_sensitive(has_account)


		def fill_account_details():
			(account_type, username, password) = self.account.get()
			builder.get_object("no_account_radio").set_active(account_type == "none")
			builder.get_object("stream_account_radio").set_active(account_type == "stream")
			builder.get_object("download_account_radio").set_active(account_type == "download")

			builder.get_object("username_entry").set_text(username or "")
			builder.get_object("password_entry").set_text(password or "")

			update_sensitivity(account_type)


		def account_type_toggled(button):
			print("account type radiobutton toggled: " + button.get_name())
			account_type = {"no_account_radio": 'none', "stream_account_radio": 'stream', "download_account_radio": 'download'} 
			if button.get_active():
				self.settings['account-type'] = account_type[button.get_name()]
				update_sensitivity(account_type[button.get_name()])

		def account_details_changed(entry, event):
			username = builder.get_object("username_entry").get_text()
			password = builder.get_object("password_entry").get_text()

			if username == "" or password == "":
				print("missing something")
				return

			# should actually try a request to http://username:password@account-type.magnatune.com/
			# to check the password is correct..

			MagnatuneAccount.instance().update(username, password)

		def format_selection_changed(button):
			self.settings['format'] = self.format_list[button.get_active()]

		self.configure_callback_dic = {
			"rb_magnatune_audio_combobox_changed_cb" : format_selection_changed,
			"rb_magnatune_radio_account_toggled_cb" : account_type_toggled
		}

		builder = Gtk.Builder()
		builder.add_from_file(rb.find_plugin_file(self, "magnatune-prefs.ui"))

		dialog = builder.get_object('magnatune_vbox')

		# Set the names of the radio buttons so we can tell which one has been clicked
		for name in ("no_account_radio", "stream_account_radio", "download_account_radio"):
			builder.get_object(name).set_name(name)

		builder.get_object("audio_combobox").set_active(self.format_list.index(self.settings['format']))

		builder.connect_signals(self.configure_callback_dic)
		builder.get_object("username_entry").connect("focus-out-event", account_details_changed)
		builder.get_object("password_entry").connect("focus-out-event", account_details_changed)

		fill_account_details()
		return dialog



# -*- Mode: python; coding: utf-8; tab-width: 8; indent-tabs-mode: t; -*-
#
# Copyright (C) 2006 Adam Zimmerman  <adam_zimmerman@sfu.ca>
# Copyright (C) 2006 James Livingston  <doclivingston@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The Rhythmbox authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and Rhythmbox. This permission is above and beyond the permissions granted
# by the GPL license by which Rhythmbox is covered. If you modify this code
# you may extend this exception to your version of the code, but you are not
# obligated to do so. If you do not wish to do so, delete this exception
# statement from your version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.

import sys
import xml.sax, xml.sax.handler
import datetime, re

import rb
from gi.repository import RB

class TrackListHandler(xml.sax.handler.ContentHandler):
	def __init__(self, db, entry_type, sku_dict, home_dict, art_dict):
		xml.sax.handler.ContentHandler.__init__(self)
		self.__db = db
		self.__entry_type = entry_type
		self.__sku_dict = sku_dict
		self.__home_dict = home_dict
		self.__art_dict = art_dict
		self.__track = {}

	def startElement(self, name, attrs):
		self.__text = ""

	def endElement(self, name):
		if name == "Track":
			try:
				# prefer ogg streams to mp3
				if 'oggurl' in self.__track:
					trackurl = self.__track['oggurl']
				else:
					trackurl = self.__track['url']

				trackurl = str(trackurl)

				# add the track to the source
				entry = self.__db.entry_lookup_by_location (trackurl)
				if entry == None:
					entry = RB.RhythmDBEntry.new(self.__db, self.__entry_type, trackurl)

				# if year is not set, use launch date instead
				try:
					year = parse_int(self.__track['year'])
					if year <= 0:
						raise ValueError
				except ValueError:
					year = parse_int(self.__track['launchdate'][0:4])

				date = datetime.date(year, 1, 1).toordinal()
				try:
					tracknum = parse_int(self.__track['tracknum'])
				except ValueError:
					tracknum = 0
				try:
					duration = parse_int(self.__track['seconds'])
				except ValueError:
					duration = 0

				self.__db.entry_set(entry, RB.RhythmDBPropType.ARTIST, str(self.__track['artist']))
				self.__db.entry_set(entry, RB.RhythmDBPropType.ALBUM, str(self.__track['albumname']))
				self.__db.entry_set(entry, RB.RhythmDBPropType.TITLE, str(self.__track['trackname']))
				self.__db.entry_set(entry, RB.RhythmDBPropType.GENRE, str(self.__track['magnatunegenres']))
				self.__db.entry_set(entry, RB.RhythmDBPropType.TRACK_NUMBER, int(tracknum))
				self.__db.entry_set(entry, RB.RhythmDBPropType.DATE, int(date))
				self.__db.entry_set(entry, RB.RhythmDBPropType.DURATION, int(duration))

				key = str(trackurl)
				sku = sys.intern(str(self.__track['albumsku']))
				self.__sku_dict[key] = sku
				self.__home_dict[sku] = str(self.__track['home'])
				self.__art_dict[sku] = str(self.__track['cover_small'])

				self.__db.commit()
			except Exception as e: # This happens on duplicate uris being added
				sys.excepthook(*sys.exc_info())
				print("Couldn't add %s - %s" % (self.__track['artist'], self.__track['trackname']), e)

			self.__track = {}
		elif name == "AllSongs":
			pass # end of the file
		else:
			self.__track[name] = self.__text

	def characters(self, content):
		self.__text = self.__text + content

# parses partial integers
def parse_int(s):
	news = ""
	for c in s:
		if c in '0123456789': # only positive integers allowed
			news += c
		else:
			break
	return int(news)



# -*- Mode: python; coding: utf-8; tab-width: 8; indent-tabs-mode: t; -*-
#
# Copyright (C) 2014  Jonathan Matthew
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The Rhythmbox authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and Rhythmbox. This permission is above and beyond the permissions granted
# by the GPL license by which Rhythmbox is covered. If you modify this code
# you may extend this exception to your version of the code, but you are not
# obligated to do so. If you do not wish to do so, delete this exception
# statement from your version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.

from gi.repository import RB, Gtk

class AlbumArtPage(object):
	def __init__(self, shell, song_info):
		self.visible = False
		self.art_key = None

		self.shell = shell
		self.song_info = song_info
		self.entry = song_info.props.current_entry
		self.art_store = RB.ExtDB(name="album-art")
		self.art_store.connect("added", self.art_added_cb)

		grid = Gtk.Grid(hexpand=True, vexpand=True, margin=6, row_spacing=6)

		self.image = RB.FadingImage(fallback="rhythmbox-missing-artwork", use_tooltip=False)
		self.image.props.hexpand = True
		self.image.props.vexpand = True
		grid.attach(self.image, 0, 0, 1, 1)

		buttons = Gtk.ButtonBox(orientation=Gtk.Orientation.HORIZONTAL)
		buttons.set_spacing(6)
		buttons.set_layout(Gtk.ButtonBoxStyle.CENTER)
		grid.attach(buttons, 0, 1, 1, 1)

		clear = Gtk.Button(label=_("Clear"), use_underline=True)
		clear.connect('clicked', self.clear_button_cb)
		buttons.add(clear)

		fetch = Gtk.Button(label=_("_Fetch"), use_underline=True)
		fetch.connect('clicked', self.fetch_button_cb)
		buttons.add(fetch)

		browse_file = Gtk.Button(label=_("_Browse"), use_underline=True)
		browse_file.connect('clicked', self.browse_button_cb)
		buttons.add(browse_file)

		self.page_num = song_info.append_page(_("Album Art"), grid)

		self.ec_id = song_info.connect("notify::current-entry", self.entry_changed_cb)
		self.sp_id = grid.get_parent().connect("switch-page", self.switch_page_cb)

	def art_update(self, key, data):
		db = self.shell.props.db
		if db.entry_matches_ext_db_key(self.entry, key):
			self.image.set_pixbuf(data)
			self.art_key = key

	def art_added_cb(self, db, key, filename, data):
		print("art added?")
		self.art_update(key, data)

	def art_request_cb(self, key, skey, filename, data):
		print("art request finished?")
		self.art_update(skey, data)

	def get_art(self, entry, user_explicit=False):
		self.image.start(100)
		key = entry.create_ext_db_key(RB.RhythmDBPropType.ALBUM)
		if user_explicit:
			key.add_info("user-explicit", "true")
		self.art_store.request(key, self.art_request_cb)

	def storage_key(self, entry):
		key = RB.ExtDBKey.create_storage("album", entry.get_string(RB.RhythmDBPropType.ALBUM))
		artist = entry.get_string(RB.RhythmDBPropType.ALBUM_ARTIST)
		if artist is None or artist == "" or artist == _("Unknown"):
			artist = entry.get_string(RB.RhythmDBPropType.ARTIST)
		key.add_field("artist", artist)
		return key

	def entry_changed_cb(self, pspec, duh):
		self.entry = self.song_info.props.current_entry

		db = self.shell.props.db
		if self.art_key and db.entry_matches_ext_db_key(self.entry, self.art_key):
			return

		self.art_key = None
		if self.visible:
			self.get_art(self.entry, False)

	def switch_page_cb(self, notebook, page, page_num):
		if self.art_key is not None:
			return

		self.visible = (page_num == self.page_num)
		if self.visible:
			self.get_art(self.entry, False)

	def clear_button_cb(self, button):
		key = self.storage_key(self.entry)
		self.art_store.store(key, RB.ExtDBSourceType.USER_EXPLICIT, None)

	def fetch_button_cb(self, button):
		if self.art_key is not None:
			self.art_store.delete(self.art_key)
		self.get_art(self.entry, True)

	def browse_file_response_cb(self, dialog, response):
		if response == Gtk.ResponseType.OK:
			key = self.storage_key(self.entry)
			self.art_store.store_uri(key, RB.ExtDBSourceType.USER_EXPLICIT, dialog.get_uri())

		dialog.destroy()

	def browse_button_cb(self, button):
		d = Gtk.FileChooserDialog(_("Select new artwork"), self.shell.props.window, Gtk.FileChooserAction.OPEN)
		d.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
		d.add_button(_("_Select"), Gtk.ResponseType.OK)
		d.set_default_response(Gtk.ResponseType.OK)
		d.connect("response", self.browse_file_response_cb)
		d.show_all()



# -*- Mode: python; coding: utf-8; tab-width: 8; indent-tabs-mode: t; -*-
#
# Copyright (C) 2009 Jonathan Matthew
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The Rhythmbox authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and Rhythmbox. This permission is above and beyond the permissions granted
# by the GPL license by which Rhythmbox is covered. If you modify this code
# you may extend this exception to your version of the code, but you are not
# obligated to do so. If you do not wish to do so, delete this exception
# statement from your version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.


#from AstrawebParser import AstrawebParser
#from WinampcnParser import WinampcnParser
from TerraParser import TerraParser
from DarkLyricsParser import DarkLyricsParser
from JlyricParser import JlyricParser
from JetlyricsParser import JetlyricsParser

from gi.repository import RB

import gettext
gettext.install('rhythmbox', RB.locale_dir())

lyrics_sites = [
	#{ 'id': 'astraweb.com', 	'class': AstrawebParser, 	'name': _("Astraweb (www.astraweb.com)") 	},
	#{ 'id': 'winampcn.com', 	'class': WinampcnParser, 	'name': _("WinampCN (www.winampcn.com)") 	},
	{ 'id': 'terra.com.br',		'class': TerraParser,		'name': _("TerraBrasil (terra.com.br)")		},
	{ 'id': 'darklyrics.com',	'class': DarkLyricsParser,	'name': _("Dark Lyrics (darklyrics.com)")	},
	{ 'id': 'j-lyric.net', 		'class': JlyricParser, 		'name': _("Jlyric (j-lyric.net)") 		},
	{ 'id': 'jetlyrics.com', 	'class': JetlyricsParser, 	'name': _("Jetlyrics (jetlyrics.com)") 		}
]




# -*- Mode: python; coding: utf-8; tab-width: 8; indent-tabs-mode: t; -*-
#
# Copyright (C) 2011 Jonathan Matthew
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The Rhythmbox authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and Rhythmbox. This permission is above and beyond the permissions granted
# by the GPL license by which Rhythmbox is covered. If you modify this code
# you may extend this exception to your version of the code, but you are not
# obligated to do so. If you do not wish to do so, delete this exception
# statement from your version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.

from gi.repository import GObject, Peas, RB, GdkPixbuf

import gettext
gettext.install('rhythmbox', RB.locale_dir())

from songinfo import AlbumArtPage

import oldcache
from lastfm import LastFMSearch
from local import LocalSearch
from musicbrainz import MusicBrainzSearch
from embedded import EmbeddedSearch

class Search(object):
	def __init__(self, store, key, last_time, searches):
		self.store = store
		self.key = key.copy()
		self.last_time = last_time
		self.searches = searches

	def next_search(self):
		if len(self.searches) == 0:
			album = self.key.get_field("album")
			if album is not None:
				key = RB.ExtDBKey.create_storage("album", album)
				key.add_field("artist", self.key.get_field("artist"))
				self.store.store(key, RB.ExtDBSourceType.NONE, None)
			return False

		search = self.searches.pop(0)
		search.search(self.key, self.last_time, self.store, self.search_done, None)
		return True

	def search_done(self, args):
		self.next_search()

class ArtSearchPlugin (GObject.GObject, Peas.Activatable):
	__gtype_name__ = 'ArtSearchPlugin'
	object = GObject.property(type=GObject.GObject)

	def __init__ (self):
		GObject.GObject.__init__ (self)

	def do_activate (self):
		self.art_store = RB.ExtDB(name="album-art")
		self.req_id = self.art_store.connect("request", self.album_art_requested)

		shell = self.object
		self.csi_id = shell.connect("create_song_info", self.create_song_info)

	def do_deactivate (self):
		self.art_store.disconnect(self.req_id)
		self.req_id = 0
		self.art_store = None

		shell = self.object
		shell.disconnect(self.csi_id)
		self.csi_id = 0

	def album_art_requested(self, store, key, last_time):
		searches = []
		if oldcache.USEFUL:
			searches.append(oldcache.OldCacheSearch())
		searches.append(EmbeddedSearch())
		searches.append(LocalSearch())
		searches.append(MusicBrainzSearch())
		searches.append(LastFMSearch())

		s = Search(store, key, last_time, searches)
		return s.next_search()

	def create_song_info(self, shell, song_info, is_multiple):
		if is_multiple is False:
			x = AlbumArtPage(shell, song_info)



# Copyright (c) 2018, 2022 Philipp Wolfer <ph.wolfer@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import json
import logging
import ssl
import time
from http.client import HTTPSConnection

HOST_NAME = "api.listenbrainz.org"
PATH_SUBMIT = "/1/submit-listens"
SSL_CONTEXT = ssl.create_default_context()


class Track:
    """
    Represents a single track to submit.

    See https://listenbrainz.readthedocs.io/en/latest/users/json.html
    """
    def __init__(self, artist_name, track_name,
                 release_name=None, additional_info={}):
        """
        Create a new Track instance
        @param artist_name as str
        @param track_name as str
        @param release_name as str
        @param additional_info as dict
        """
        self.artist_name = artist_name
        self.track_name = track_name
        self.release_name = release_name
        self.additional_info = additional_info

    @staticmethod
    def from_dict(data):
        return Track(
            data["artist_name"],
            data["track_name"],
            data.get("release_name", None),
            data.get("additional_info", {})
        )

    def to_dict(self):
        return {
            "artist_name": self.artist_name,
            "track_name": self.track_name,
            "release_name": self.release_name,
            "additional_info": self.additional_info
        }

    def __repr__(self):
        return "Track(%s, %s)" % (self.artist_name, self.track_name)


class ListenBrainzClient:
    """
    Submit listens to ListenBrainz.org.

    See https://listenbrainz.readthedocs.io/en/latest/users/api/index.html
    """

    def __init__(self, logger=logging.getLogger(__name__), timeout=30):
        self.__next_request_time = 0
        self.user_token = None
        self.logger = logger
        self._timeout = timeout

    def listen(self, listened_at, track):
        """
        Submit a listen for a track
        @param listened_at as int
        @param entry as Track
        """
        payload = _get_payload(track, listened_at)
        return self._submit("single", [payload])

    def playing_now(self, track):
        """
        Submit a playing now notification for a track
        @param track as Track
        """
        payload = _get_payload(track)
        return self._submit("playing_now", [payload])

    def import_tracks(self, tracks):
        """
        Import a list of tracks as (listened_at, Track) pairs
        @param track as [(int, Track)]
        """
        payload = _get_payload_many(tracks)
        return self._submit("import", payload)

    def _submit(self, listen_type, payload, retry=0):
        self._wait_for_ratelimit()
        self.logger.debug("ListenBrainz %s: %r", listen_type, payload)
        data = {
            "listen_type": listen_type,
            "payload": payload
        }
        headers = {
            "Authorization": "Token %s" % self.user_token,
            "Content-Type": "application/json"
        }
        body = json.dumps(data)
        conn = HTTPSConnection(HOST_NAME, context=SSL_CONTEXT, timeout=self._timeout)
        try:
            conn.request("POST", PATH_SUBMIT, body, headers)
            response = conn.getresponse()
            response_text = response.read()
        finally:
            conn.close()

        try:
            response_data = json.loads(response_text)
        except json.decoder.JSONDecodeError:
            response_data = response_text

        self._handle_ratelimit(response)
        log_msg = "Response %s: %r" % (response.status, response_data)
        if response.status == 429 and retry < 5:  # Too Many Requests
            self.logger.warning(log_msg)
            return self._submit(listen_type, payload, retry + 1)
        elif response.status == 200:
            self.logger.debug(log_msg)
        else:
            self.logger.error(log_msg)
        return response

    def _wait_for_ratelimit(self):
        now = time.time()
        if self.__next_request_time > now:
            delay = self.__next_request_time - now
            self.logger.debug("Rate limit applies, delay %d", delay)
            time.sleep(delay)

    def _handle_ratelimit(self, response):
        remaining = int(response.getheader("X-RateLimit-Remaining", 0))
        reset_in = int(response.getheader("X-RateLimit-Reset-In", 0))
        self.logger.debug("X-RateLimit-Remaining: %i", remaining)
        self.logger.debug("X-RateLimit-Reset-In: %i", reset_in)
        if remaining == 0:
            self.__next_request_time = time.time() + reset_in


def _get_payload_many(tracks):
    payload = []
    for (listened_at, track) in tracks:
        data = _get_payload(track, listened_at)
        payload.append(data)
    return payload


def _get_payload(track, listened_at=None):
    data = {
        "track_metadata": track.to_dict()
    }
    if listened_at is not None:
        data["listened_at"] = listened_at
    return data



# Copyright (c) 2018-2019, 2022-2023 Philipp Wolfer <ph.wolfer@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import logging
import re
import sys
import threading
import time
from gi.repository import GObject
from gi.repository import Peas
from gi.repository import RB
from client import ListenBrainzClient, Track
from queue import ListenBrainzQueue
from settings import ListenBrainzSettings, load_settings


UNSUPPORTED_TYPES = ('podcast-post')
RE_MBID = re.compile(r"[0-9a-f]{8}(?:-[0-9a-f]{4}){3}-[0-9a-f]{12}")


logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
logger = logging.getLogger("listenbrainz")


class ListenBrainzPlugin(GObject.Object, Peas.Activatable):
    __gtype_name = 'ListenBrainzPlugin'
    object = GObject.property(type=GObject.GObject)

    def __init__(self):
        GObject.Object.__init__(self)
        self.settings = None
        self.__client = None
        self.__current_entry = None
        self.__current_start_time = 0
        self.__current_elapsed = 0
        self.__lock = threading.Lock()

    def do_activate(self):
        logger.debug("activating ListenBrainz plugin")
        self.settings = load_settings()
        self.__client = ListenBrainzClient(logger=logger)
        self.settings.connect("changed::user-token",
                              self.on_user_token_changed)
        self.on_user_token_changed(self.settings)
        self.__current_entry = None
        self.__current_start_time = 0
        self.__current_elapsed = 0
        self.__queue = ListenBrainzQueue(self.__client)
        with self.__lock:
            try:
                self.__queue.load()
            except Exception as e:
                _handle_exception(e)
        self.__queue.activate()
        shell_player = self.object.props.shell_player
        shell_player.connect("playing-song-changed",
                             self.on_playing_song_changed)
        shell_player.connect("elapsed-changed", self.on_elapsed_changed)

    def do_deactivate(self):
        logger.debug("deactivating ListenBrainz plugin")
        shell_player = self.object.props.shell_player
        shell_player.disconnect_by_func(self.on_playing_song_changed)
        shell_player.disconnect_by_func(self.on_elapsed_changed)
        self.settings.disconnect_by_func(self.on_user_token_changed)
        self.__queue.deactivate()
        with self.__lock:
            self.__queue.submit_batch()
            self.__queue.save()

    def on_playing_song_changed(self, player, entry):
        logger.debug("playing-song-changed: %r, %r", player, entry)

        self._submit_current_entry()

        self.__current_entry = entry
        self.__current_elapsed = 0

        if not _can_be_listened(entry):
            self.__current_entry = None
            return

        self.__current_start_time = int(time.time())
        track = _entry_to_track(entry)
        threading.Thread(
            target=self._run_in_thread,
            args=[self.__client.playing_now, track]
            ).start()

    def on_elapsed_changed(self, player, elapsed):
        # logger.debug("elapsed-changed: %r, %i" % (player, elapsed))
        if player.get_playing_entry() == self.__current_entry:
            self.__current_elapsed += 1

    def on_user_token_changed(self, settings, key="user-token"):
        self.__client.user_token = settings.get_string("user-token")

    def _submit_current_entry(self):
        if self.__current_entry is not None:
            duration = self.__current_entry.get_ulong(
                            RB.RhythmDBPropType.DURATION)
            elapsed = self.__current_elapsed
            logger.debug("Elapsed: %s / %s", elapsed, duration)
            if elapsed >= 240 or (duration and elapsed >= duration / 2):
                track = _entry_to_track(self.__current_entry)
                threading.Thread(
                    target=self._run_in_thread,
                    args=[self.__queue.add, self.__current_start_time, track]
                    ).start()

    def _run_in_thread(self, callable, *args):
        with self.__lock:
            try:
                callable(*args)
            except Exception as e:
                _handle_exception(e)


def _can_be_listened(entry):
    if entry is None:
        return False

    entry_type = entry.get_entry_type()
    category = entry_type.get_property("category")
    title = entry.get_string(RB.RhythmDBPropType.TITLE)
    error = entry.get_string(RB.RhythmDBPropType.PLAYBACK_ERROR)

    if category != RB.RhythmDBEntryCategory.NORMAL:
        logger.debug('Cannot submit %r: Category "%s"' %
                     (title, category.value_name))
        return False

    type_name = entry_type.get_name()
    if type_name in UNSUPPORTED_TYPES or _is_unidentified_audiocd(entry):
        logger.debug('Cannot submit listen %r: Entry type "%s"' %
                     (title, type_name))
        return False

    if error is not None:
        logger.debug("Cannot submit %r: Playback error %s" %
                     (entry.get_string(RB.RhythmDBPropType.LOCATION), error))
        return False

    return True


def _is_unidentified_audiocd(entry):
    entry_type = entry.get_entry_type()
    return (entry_type.get_name().startswith("audiocd")
            and not entry.get_string(RB.RhythmDBPropType.MB_ALBUMID))


def _handle_exception(e):
    logger.error("ListenBrainz exception %s: %s", type(e).__name__, e)


def _entry_to_track(entry):
    artist = entry.get_string(RB.RhythmDBPropType.ARTIST)
    title = entry.get_string(RB.RhythmDBPropType.TITLE)
    album = entry.get_string(RB.RhythmDBPropType.ALBUM)
    track_number = entry.get_ulong(RB.RhythmDBPropType.TRACK_NUMBER)
    mb_track_id = entry.get_string(RB.RhythmDBPropType.MB_TRACKID)
    mb_album_id = entry.get_string(RB.RhythmDBPropType.MB_ALBUMID)
    mb_artist_id = entry.get_string(RB.RhythmDBPropType.MB_ARTISTID)
    duration = entry.get_ulong(RB.RhythmDBPropType.DURATION)

    additional_info = {
        "release_mbid": _validate_mbid(mb_album_id),
        "recording_mbid": _validate_mbid(mb_track_id),
        "artist_mbids": [mb_artist_id] if _validate_mbid(mb_artist_id) else [],
        "tracknumber": track_number or None,
        "duration": duration,
        "media_player": "Rhythmbox",
        "submission_client": "ListenBrainz plugin",
    }

    entry_type = entry.get_entry_type().get_name()
    if entry_type != "song" and not entry_type.startswith("audiocd"):
        additional_info["music_service_name"] = _cleanup_source(entry_type)

    return Track(artist, title, album, additional_info)


def _validate_mbid(mbid):
    return mbid if RE_MBID.match(mbid) else None


def _cleanup_source(source):
    if source.startswith("grilo:grl-"):
        return source[10:]
    return source


GObject.type_register(ListenBrainzSettings)



# -*- Mode: python; coding: utf-8; tab-width: 8; indent-tabs-mode: t; -*-
#
# Copyright (C) 2007 James Livingston
# Copyright (C) 2007 Sirio Bolaños Puchet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The Rhythmbox authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and Rhythmbox. This permission is above and beyond the permissions granted
# by the GPL license by which Rhythmbox is covered. If you modify this code
# you may extend this exception to your version of the code, but you are not
# obligated to do so. If you do not wish to do so, delete this exception
# statement from your version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.

from LyricsSites import lyrics_sites

from os import system, path

import rb
from gi.repository import RB, Gtk, Gio, GObject, PeasGtk

import gettext
gettext.install('rhythmbox', RB.locale_dir())

class LyricsConfigureDialog (GObject.Object, PeasGtk.Configurable):
	__gtype_name__ = 'LyricsConfigureDialog'
	object = GObject.property(type=GObject.Object)

	def __init__(self):
		GObject.Object.__init__(self)
		self.settings = Gio.Settings.new("org.gnome.rhythmbox.plugins.lyrics")

	def do_create_configure_widget(self):
		builder = Gtk.Builder()
		builder.add_from_file(rb.find_plugin_file(self, "lyrics-prefs.ui"))

		self.config = builder.get_object("config")

		self.choose_button = builder.get_object("choose_button")
		self.path_display = builder.get_object("path_display")

		self.choose_button.connect("clicked", self.choose_callback)

		engines, self.folder = self.get_prefs()
		if self.folder is None:
			self.folder = '~/.lyrics'
		self.path_display.set_text(self.folder)

		# build site list
		site_box = builder.get_object("sites")
		self.site_checks = {}
		for s in lyrics_sites:
			site_id = s['id']
			checkbutton = Gtk.CheckButton(label = s['name'])
			checkbutton.set_active(s['id'] in engines)
			checkbutton.connect("toggled", self.set_sites)
			self.site_checks[site_id] = checkbutton
			site_box.pack_start(checkbutton, True, True, 0)

		site_box.show_all()

		return self.config

	def set_sites(self, widget):
		sites = []
		for s in lyrics_sites:
			check = self.site_checks[s['id']]
			if check is None:
				continue

			if check.get_active():
				sites.append(s['id'])

		print("setting lyrics sites: " + str(sites))
		self.settings['sites'] = sites


	def choose_callback(self, widget):
		def response_handler(widget, response):
			if response == Gtk.ResponseType.OK:
				path = self.chooser.get_filename()
				self.chooser.destroy()
				self.path_display.set_text(path)
				self.settings['folder'] = path
			else:
				self.chooser.destroy()

		buttons = (Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE,
				Gtk.STOCK_OK, Gtk.ResponseType.OK)
		self.chooser = Gtk.FileChooserDialog(title=_("Choose lyrics folder..."),
					parent=None,
					action=Gtk.FileChooserAction.SELECT_FOLDER,
					buttons=buttons)
		self.chooser.connect("response", response_handler)
		self.chooser.set_modal(True)
		self.chooser.set_transient_for(self.config.get_toplevel())
		self.chooser.present()

	def get_prefs (self):
		try:
			sites = self.settings['sites']
		except GLib.GError as e:
			print(e)
			engines = []
		folder = self.settings['folder']

		print("lyric sites: " + str (sites))
		print("lyric folder: " + folder)
		return (sites, folder)



# -*- Mode: python; coding: utf-8; tab-width: 8; indent-tabs-mode: t; -*-
#
# Copyright (C) 2007 Jonathan Matthew
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The Rhythmbox authors hereby grant permission for non-GPL compatible
# GStreamer plugins to be used and distributed together with GStreamer
# and Rhythmbox. This permission is above and beyond the permissions granted
# by the GPL license by which Rhythmbox is covered. If you modify this code
# you may extend this exception to your version of the code, but you are not
# obligated to do so. If you do not wish to do so, delete this exception
# statement from your version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.


import urllib.parse
import rb
from xml.dom import minidom

class LyricWikiParser(object):
	def __init__(self, artist, title):
		self.artist = artist
		self.title = title
	
	def search(self, callback, *data):
		artist = urllib.parse.quote(self.artist.replace(' ', '_'))
		title = urllib.parse.quote(self.title.replace(' ', '_'))

		htstring = 'http://lyricwiki.org/api.php?artist=%s&song=%s&fmt=text' % (artist, title)
			
		loader = rb.Loader()
		loader.get_url (htstring, self.got_lyrics, callback, *data)

	def got_lyrics(self, result, callback, *data):
		if result is None or result == "Not found":
			callback (None, *data)
			return

		result = result.decode('iso-8859-1')
		result += "\n\nLyrics provided by lyricwiki.org"

		callback (result, *data)
